// ==============================================================
//                 ORBITER MODULE: Atlantis
//                  Part of the ORBITER SDK
//            Copyright (C) 2001 Martin Schweiger
//                   All rights reserved
//
// Atlantis.cpp
// Reference implementation of Atlantis (Space Shuttle) vessel
// class module
// ==============================================================

#include "Atlantis.h"
#include <stdio.h>
#include <fstream.h>

// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,   -0.2,     0,   0.8,     1,   0.8,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	for (int i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
	return CL[i] + (aoa-AOA[i])*SCL[i];
}

extern void GetSRB_State (double srb_time, double &thrust_level, double &prop_level);

// ==============================================================
// Specialised vessel class Atlantis
// ==============================================================

// Constructor
Atlantis::Atlantis (OBJHANDLE hObj, int fmodel)
: VESSEL(hObj, fmodel)
{
	status          = 0;
	srb_thrust      = 0;
	gear_status     = DOOR_CLOSED;
	gear_proc       = 0.0;
	gear_moving     = false;
	gear_endanim    = false;
	door_status     = DOOR_CLOSED;
	door_proc       = 0.0;
	door_moving     = false;
	door_endanim    = false;
	bManualSeparate = false;
	vis             = NULL;

	// preload meshes
	hOrbiterMesh = oapiLoadMeshGlobal ("Atlantis");
	hTankMesh    = oapiLoadMeshGlobal ("Atlantis_tank");
	hSRBMesh     = oapiLoadMeshGlobal ("Atlantis_srb");
}

// Read status from scenario file
void Atlantis::LoadState (FILEHANDLE scn, VESSELSTATUS *vs)
{
    char *line;
	double srbtime = 0.0;

	while (oapiReadScenario_nextline (scn, line)) {
        if (!strnicmp (line, "CONFIGURATION", 13)) {
            sscanf (line+13, "%d", &status);
		} else if (!strnicmp (line, "CARGODOOR", 9)) {
			sscanf (line+9, "%d%lf", &door_status, &door_proc);
		} else if (!strnicmp (line, "GEAR", 4)) {
			sscanf (line+4, "%d%lf", &gear_status, &gear_proc);
		} else if (!strnicmp (line, "SRB_IGNITION_TIME", 17)) {
			sscanf (line+17, "%lf", &srbtime);
        } else {
            ParseScenarioLine (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
    }
	switch (status) {
	case 0:
		SetLaunchConfiguration();
		break;
	case 1:
		SetPostLaunchConfiguration (-srbtime); // assuming simulation starts at t=0
		break;
	case 2:
		SetOrbiterTankConfiguration();
		break;
	case 3:
		SetOrbiterConfiguration();
		break;
	}

	// launch animations if required
	if (gear_status == DOOR_OPENING || gear_status == DOOR_CLOSING)
		OperateLandingGear();
	if (door_status == DOOR_OPENING || door_status == DOOR_CLOSING)
		OperateCargoDoors();
}

// Write status to scenario file
void Atlantis::SaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	SaveDefaultState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "CONFIGURATION", status);

	if (status == 1)
		oapiWriteScenario_float (scn, "SRB_IGNITION_TIME", oapiGetSimTime()-srb_ignition_time);

	sprintf (cbuf, "%d %0.4f", door_status, door_proc);
	oapiWriteScenario_string (scn, "CARGODOOR", cbuf);

	sprintf (cbuf, "%d %0.4f", gear_status, gear_proc);
	oapiWriteScenario_string (scn, "GEAR", cbuf);
}

// launch configuration setup (orbiter + tank + SRB)
void Atlantis::SetLaunchConfiguration (void)
{
	// Physical parameters
	SetSize (30.0);

	SetEmptyMass (ORBITER_EMPTY_MASS + ORBITER_MAX_PROPELLANT_MASS +
				  TANK_EMPTY_MASS +
				  2 * (SRB_EMPTY_MASS + SRB_MAX_PROPELLANT_MASS));
	// Note:
	// 1. During launch the orbiter burns hyrdogen/oxygen from the main
	//    tank. The orbiter's internal fuel supply is not touched yet,
	//    so is added to the 'empty' mass.
	// 2. SRB fuel consumption and thrust are handled outside the orbiter
	//    core, so the SRB propellant mass is also added to the 'empty' mass.

	SetMaxFuelMass (719115.0);
	// Tank propellant

	SetISP (5000.0);
	// Indirectly derived from fuel mass and burn time,
	// realistic value should be around 4255

	SetMaxThrust (ENGINE_MAIN, 5005956.0);
	// 1668652 N * 3 engines
	// Main engines only (SRB's handled separately)
	// Note that this is the nominal thrust at sea level.
	// In vacuum this should increase to 2091374 N per engine

	SetMaxThrust (ENGINE_RETRO, 0);
	SetMaxThrust (ENGINE_HOVER, 0);

	SetMaxThrust (ENGINE_ATTITUDE, 5e4);
	// At this stage attitude control  is performed by thrust-vectoring
	// The RCS (Reaction Control System) is activated only after tank separation
	// The above thrust value is speculative

	SetCOG_elev (53.3);
	SetCW (0.2, 0.5, 1.5, 1.5);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(687.4,849.5,189.4));
	SetRotDrag (_V(0.7,0.1,0.3));
	SetPMI (_V(179.1,176.8,29.3));
	SetTrimScale (0.05);
	SetLiftCoeffFunc (0); // simplification: we assume no lift during launch phase
	SetDockParams (ORBITER_DOCKPOS+OFS_LAUNCH_ORBITER, _V(0,1,0), _V(0,0,-1));
	SetTouchdownPoints (_V(0,-10,-53.3), _V(-7,7,-53.3), _V(7,7,-53.3));
	srb_thrust = GetMainThrustModPtr();

	// Visual parameters
	ClearMeshes();
	ClearExhaustRefs();
	ClearAttExhaustRefs();
	AddOrbiterVisual (OFS_LAUNCH_ORBITER, true, false);
	AddTankVisual    (OFS_LAUNCH_TANK);
	AddSRBVisual     (OFS_LAUNCH_RIGHTSRB);
	AddSRBVisual     (OFS_LAUNCH_LEFTSRB);

	status = 0;
}

void Atlantis::SetPostLaunchConfiguration (double srbtime)
{
	SetLaunchConfiguration();
	srb_ignition_time = srbtime;
	status = 1;
}

// Configuration after booster separation (orbiter + tank)
void Atlantis::SetOrbiterTankConfiguration (void)
{
	// Physical parameters
	SetSize (28.8);
	SetEmptyMass (ORBITER_EMPTY_MASS + ORBITER_MAX_PROPELLANT_MASS + TANK_EMPTY_MASS);
	SetMaxThrust (ENGINE_MAIN, 6274122.0); // the vacuum value
	SetMaxThrust (ENGINE_RETRO, 0);
	SetMaxThrust (ENGINE_HOVER, 0);
	SetMaxThrust (ENGINE_ATTITUDE, 5e4);
	SetCW (0.2, 0.5, 1.5, 1.5);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(646.1,603.0,141.5));
	SetRotDrag (_V(0.7,0.1,0.3));
	SetPMI (_V(173.3,161.0,24.0));
	SetTrimScale (0.05);
	SetLiftCoeffFunc (0); // simplification: we assume no lift during launch phase
	SetDockParams (ORBITER_DOCKPOS+OFS_WITHTANK_ORBITER, _V(0,1,0), _V(0,0,-1));
	//SetCOG_elev (53.3); // fix
	//SetTouchdownPoints (_V(0,-10,-53.3), _V(-7,7,-53.3), _V(7,7,-53.3)); // fix
	SetTouchdownPoints (_V(0,-5,30), _V(-10,-10,-30), _V(10,0,-30));

	// Visual parameters
	ClearMeshes();
	ClearExhaustRefs();
	ClearAttExhaustRefs();
	AddOrbiterVisual (OFS_WITHTANK_ORBITER, true, false);
	AddTankVisual    (OFS_WITHTANK_TANK);

	status = 2;
}

// Configuration after tank separation (orbiter only)
void Atlantis::SetOrbiterConfiguration (void)
{
	// Physical parameters
	SetSize (19.6);
	SetEmptyMass (ORBITER_EMPTY_MASS);
	SetMaxFuelMass (ORBITER_MAX_PROPELLANT_MASS);

	SetFuelMass (ORBITER_MAX_PROPELLANT_MASS);
	// switching to internal tanks

	SetMaxThrust (ENGINE_MAIN, ORBITER_OMS_THRUST);
	SetISP (ORBITER_OMS_ISP);
	// configure OMS (Orbital Maneuvering System)

	SetMaxThrust (ENGINE_ATTITUDE, ORBITER_RCS_THRUST);
	// configure RCS (Reaction Control System)

	SetPMI (_V(78.2,82.1,10.7));
	SetCW (ORBITER_CW[0], ORBITER_CW[1], ORBITER_CW[2], ORBITER_CW[3]);
	SetCrossSections (ORBITER_CS);
	SetCOG_elev (8.0);
	if (gear_status == DOOR_OPEN)
		SetTouchdownPoints (_V(0,-4,17), _V(-3.96,-5,-4.3), _V(3.96,-5,-4.3)); // gear wheel tips
	else
		SetTouchdownPoints (_V(0,-2.5,14), _V(-8,-2.8,-9), _V(8,-2.8,-9)); // belly landing
	SetLiftCoeffFunc (LiftCoeff);

	if (GetFlightModel() >= 1) { // realistic flight model
		SetPitchMomentScale (1e-5);
		SetBankMomentScale (2e-5);
		if (gear_status == DOOR_OPEN) {
			SetCW (ORBITER_CW[0]+ORBITER_CW_GEAR[0],
				   ORBITER_CW[1]+ORBITER_CW_GEAR[1],
				   ORBITER_CW[2]+ORBITER_CW_GEAR[2],
				   ORBITER_CW[3]+ORBITER_CW_GEAR[3]);
			SetCrossSections (ORBITER_CS+ORBITER_CS_GEAR);
		}
	}

	SetRotDrag (_V(0.5,1.0,1.0));

	SetDockParams (ORBITER_DOCKPOS, _V(0,1,0), _V(0,0,-1));
	// for docking attachment in payload bay

	// Visual parameters
	ClearMeshes();
	ClearExhaustRefs();
	ClearAttExhaustRefs();
	AddOrbiterVisual (OFS_ZERO, false, true);

	status = 3;
}

void Atlantis::AddOrbiterVisual (const VECTOR3 &ofs, bool maineng, bool atteng)
{
	AddMesh (hOrbiterMesh, &ofs);
	SetCameraOffset (_V(ofs.x,ofs.y+2.4,ofs.z+14));
	VECTOR3 dir;

	if (maineng) {
		dir = _V(-0.087, 0.0, -0.996);
		AddExhaustRef (EXHAUST_MAIN, _V(-1.6+ofs.x,-0.2+ofs.y,-16.0+ofs.z), 8.0, 1.0, &dir);
		dir.x = 0.087;
		AddExhaustRef (EXHAUST_MAIN, _V( 1.6+ofs.x,-0.2+ofs.y,-16.0+ofs.z), 8.0, 1.0, &dir);
		dir.x = 0.0, dir.y = 0.17, dir.z = -0.985;
		AddExhaustRef (EXHAUST_MAIN, _V(     ofs.x, 2.7+ofs.y,-16.0+ofs.z), 8.0, 1.0, &dir);
	} else { // set up orbital maneuvering system
		dir = _V(-0.050, 0.099, -0.994);
		AddExhaustRef (EXHAUST_MAIN, _V(-2.05,3.45,-14.2)+ofs, 4.0, 0.5, &dir);
		dir.x = 0.050;
		AddExhaustRef (EXHAUST_MAIN, _V( 2.05,3.45,-14.2)+ofs, 4.0, 0.5, &dir);
	}
	if (atteng) {
		// add attitude render definitions
		UINT idx;
		idx = AddAttExhaustRef (_V(0,1.6,17)+ofs, _V(0,1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
		idx = AddAttExhaustRef (_V(-0.4,1.55,17)+ofs, _V(0,1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
		idx = AddAttExhaustRef (_V(0.4,1.55,17)+ofs, _V(0,1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
		idx = AddAttExhaustRef (_V(-3.4,1.6,-14.4)+ofs, _V(0,-1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
		idx = AddAttExhaustRef (_V(3.4,1.6,-14.4)+ofs, _V(0,-1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
		idx = AddAttExhaustRef (_V(0,-2.5,16)+ofs, _V(0,-1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
		idx = AddAttExhaustRef (_V(-0.4,-2.5,16)+ofs, _V(0,-1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
		idx = AddAttExhaustRef (_V(0.4,-2.5,16)+ofs, _V(0,-1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
		idx = AddAttExhaustRef (_V(-3.4,3.3,-14.4)+ofs, _V(0,1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
		idx = AddAttExhaustRef (_V(3.4,3.3,-14.4)+ofs, _V(0,1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
		AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
		idx = AddAttExhaustRef (_V(-2.0, -0.6, 17.2)+ofs, _V(-0.174,-0.985,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 2, 0); // roll right
		idx = AddAttExhaustRef (_V(2.0, -0.6, 17.2)+ofs, _V(0.174,-0.985,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 2, 1); // roll left
		idx = AddAttExhaustRef (_V(1.9, -0.6, 17.6)+ofs, _V(0.174,0.985,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 2, 0); // roll right
		idx = AddAttExhaustRef (_V(-1.9, -0.6, 17.6)+ofs, _V(-0.174,0.985,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 2, 1); // roll left
		idx = AddAttExhaustRef (_V(-3.4, 1.5,-14.0)+ofs, _V(0,-1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 2, 0); // roll right
		idx = AddAttExhaustRef (_V(3.4, 1.5,-14.0)+ofs, _V(0,-1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 2, 1); // roll left
		idx = AddAttExhaustRef (_V(-3.4,3.3,-14.0)+ofs, _V(0,1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 2, 1); // roll left
		idx = AddAttExhaustRef (_V(3.4,3.3,-14.0)+ofs, _V(0,1,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 2, 0); // roll right
		idx = AddAttExhaustRef (_V(-1.9,-0.6,17.6)+ofs, _V(-1,0,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 1, 1); // yaw right
		AddAttExhaustMode (idx, ATTMODE_LIN, 0, 0); // move right
		idx = AddAttExhaustRef (_V(1.9,-0.6,17.6)+ofs, _V(1,0,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 1, 0); // yaw left
		AddAttExhaustMode (idx, ATTMODE_LIN, 0, 1); // move left
		idx = AddAttExhaustRef (_V(-4,2.5,-14.4)+ofs, _V(-1,0,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 1, 0); // yaw left
		AddAttExhaustMode (idx, ATTMODE_LIN, 0, 0); // move right
		idx = AddAttExhaustRef (_V(4,2.5,-14.4)+ofs, _V(1,0,0), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_ROT, 1, 1); // yaw right
		AddAttExhaustMode (idx, ATTMODE_LIN, 0, 1); // move left
		idx = AddAttExhaustRef (_V(0,1.2,18.2), _V(0,0,1), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_LIN, 2, 1); // move back
		idx = AddAttExhaustRef (_V(-0.4,1.15,18.2), _V(0,0,1), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_LIN, 2, 1); // move back
		idx = AddAttExhaustRef (_V(0.4,1.15,18.2), _V(0,0,1), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_LIN, 2, 1); // move back
		idx = AddAttExhaustRef (_V(-3.4,2.5,-15), _V(0,0,-1), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_LIN, 2, 0); // move forward
		idx = AddAttExhaustRef (_V(3.4, 2.5,-15), _V(0,0,-1), 0.3, 0.3);
		AddAttExhaustMode (idx, ATTMODE_LIN, 2, 0); // move forward
	}
}

void Atlantis::AddTankVisual (const VECTOR3 &ofs)
{
	UINT meshidx;
	meshidx = AddMesh (hTankMesh, &ofs);
	SetMeshVisibleInternal (meshidx, true);
}

void Atlantis::AddSRBVisual (const VECTOR3 &ofs)
{
	UINT meshidx;
	meshidx = AddMesh (hSRBMesh, &ofs);
	SetMeshVisibleInternal (meshidx, true);
	int id = AddExhaustRef (EXHAUST_CUSTOM, _V(ofs.x, ofs.y, ofs.z-21.8), 0, 0);
	if (ofs.x < 0.0) srb_id1 = id;
	else             srb_id2 = id;
}

void Atlantis::SeparateBoosters (double srb_time)
{
	// Create SRB's as individual objects
	VESSELSTATUS vs;
	GetStatus (vs);
	GetSRB_State (srb_time, vs.eng_main, vs.fuel);
	VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
	VECTOR3 vel = {3, -3, 0};
	Local2Rel (OFS_LAUNCH_RIGHTSRB, vs.rpos);
	GlobalRot (vel, rofs);
	vs.rvel.x = rvel.x+rofs.x;
	vs.rvel.y = rvel.y+rofs.y;
	vs.rvel.z = rvel.z+rofs.z;
	vs.vrot.y -= 0.1;
	vs.vrot.x -= 0.1;
	vs.status = 0;
	char name[256];
	strcpy (name, GetName()); strcat (name, "-SRB1");
	oapiCreateVessel (name, "Atlantis_SRB", vs);
	vel.x = -3;
	Local2Rel (OFS_LAUNCH_LEFTSRB, vs.rpos);
	GlobalRot (vel, rofs);
	vs.rvel.x = rvel.x+rofs.x;
	vs.rvel.y = rvel.y+rofs.y;
	vs.rvel.z = rvel.z+rofs.z;
	vs.vrot.x += 0.2;
	vs.arot.z += PI;
	name[strlen(name)-1] = '2';
	oapiCreateVessel (name, "Atlantis_SRB", vs);

	// Remove SRB's from Shuttle instance
	*srb_thrust = 0.0;
	SetOrbiterTankConfiguration();
	ShiftCentreOfMass (OFS_LAUNCH_ORBITER-OFS_WITHTANK_ORBITER);
}

void Atlantis::SeparateTank (void)
{
	// Create Tank as individual object
	VESSELSTATUS vs;
	GetStatus (vs);
	vs.eng_main = vs.eng_hovr = 0.0;
	VECTOR3 ofs = OFS_WITHTANK_TANK;
	VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
	VECTOR3 vel = {0,-1,0};
	Local2Rel (ofs, vs.rpos);
	GlobalRot (vel, rofs);
	vs.rvel.x = rvel.x+rofs.x;
	vs.rvel.y = rvel.y+rofs.y;
	vs.rvel.z = rvel.z+rofs.z;
	vs.vrot.x = -0.02;
	vs.status = 0;
	char name[256];
	strcpy (name, GetName()); strcat (name, "-Tank");
	oapiCreateVessel (name, "Atlantis_Tank", vs);

	// Remove Tank from shuttle instance
	SetOrbiterConfiguration ();
	ShiftCentreOfMass (OFS_WITHTANK_ORBITER);
}

void Atlantis::Timestep (double simt)
{
	const double eps = 1e-6;
	double m_fuel, srb_time;

	switch (status) {
	case 0: // launch configuration
		if (GetEngineLevel (ENGINE_MAIN) > 0.95) {
			status = 1; // launch
			srb_ignition_time = simt; // store SRB reference time
		}
		break;
	case 1: // SRB's ignited
		srb_time = simt-srb_ignition_time;
		if (srb_time > SRB_SEPARATION_TIME || bManualSeparate) { // separate boosters
			SeparateBoosters (srb_time);
			bManualSeparate = false;
		} else {
			// extract current thrust level and propellant level as a function of time
			double thrust_level, prop_level;
			GetSRB_State (srb_time, thrust_level, prop_level);

			// incorporate pressure-dependency of SRB thrust
			double srb_maxthrust = SRB_FREE_THRUST * exp (SRB_THRUST_PRESSURE_EXP * GetAtmPressure());

			// set SRB thrust component
			*srb_thrust = 2.0*srb_maxthrust*thrust_level;

			// set dimensions of SRB exhaust visualisation
			double exhaust_scale_1 = 16.0*thrust_level;
			double exhaust_scale_2 = 2.0*sqrt(thrust_level);
			SetExhaustScales (EXHAUST_CUSTOM, srb_id1, exhaust_scale_1, exhaust_scale_2);
			SetExhaustScales (EXHAUST_CUSTOM, srb_id2, exhaust_scale_1, exhaust_scale_2);
			
			// Since the SRB's are managed outside the Orbiter core, the SRB mass (including
			// propellant) is registered as 'empty mass'
			SetEmptyMass (ORBITER_EMPTY_MASS + ORBITER_MAX_PROPELLANT_MASS + TANK_EMPTY_MASS +
				2.0*(SRB_EMPTY_MASS + SRB_MAX_PROPELLANT_MASS*prop_level));
		}
		break;
	case 2: // post SRB separation
		m_fuel = GetMass() - GetEmptyMass();
		if (m_fuel/GetMaxFuelMass() < eps || bManualSeparate) {
			SeparateTank();
			bManualSeparate = false;
		}
		break;
	}
	if (door_moving) { // update payload bay door status
		double dt = oapiGetSimStep();
		double da = dt * DOOR_OPERATING_SPEED;
		if (door_status == DOOR_CLOSING) {
			if (door_endanim) {
				door_status = DOOR_CLOSED;
				door_moving = door_endanim = false;
				UnregisterAnimation();
			} else if ((door_proc -= da) <= 0.0) {
				door_proc   = 0.0;
				door_endanim = true;
			}
		} else if (door_status = DOOR_OPENING) {
			if (door_endanim) {
				door_status = DOOR_OPEN;
				door_moving = door_endanim = false;
				UnregisterAnimation();
			} else if ((door_proc += da) >= 1.0) {
				door_proc   = 1.0;
				door_endanim = true;
			}
		}
	}
	if (gear_moving) { // update landing gear status
		double dt = oapiGetSimStep();
		double da = dt * GEAR_OPERATING_SPEED;
		double pgear_proc = gear_proc;
		if (gear_status == DOOR_CLOSING) {
			if (gear_endanim) {
				gear_status = DOOR_CLOSED;
				gear_moving = gear_endanim = false;
				UnregisterAnimation();
			} else if ((gear_proc -= da) <= 0.0) {
				gear_proc    = 0.0;
				gear_endanim = true;
				SetTouchdownPoints (_V(0,-2.5,14), _V(-8,-2.8,-9), _V(8,-2.8,-9));
				// gear retracted: set contact points for belly landing
			}
		} else if (gear_status = DOOR_OPENING) {
			if (gear_endanim) {
				gear_status = DOOR_OPEN;
				gear_moving = gear_endanim = false;
				UnregisterAnimation();
			} else if ((gear_proc += da) >= 1.0) {
				gear_proc    = 1.0;
				gear_endanim = true;
				SetTouchdownPoints (_V(0,-4,17), _V(-3.96,-5,-4.3), _V(3.96,-5,-4.3));
				// gear lowered: set gear contact points
			}
		}
		// update aerodynamics for current gear status
		if (GetFlightModel() >= 1) {
			double delta = gear_proc-pgear_proc;
			VECTOR3 cs;
			GetCrossSections (cs);
			SetCrossSections (cs + ORBITER_CS_GEAR*delta);
			double cw[4];
			GetCW (cw[0], cw[1], cw[2], cw[3]);
			for (int i = 0; i < 4; i++) cw[i] += ORBITER_CW_GEAR[i]*delta;
			SetCW (cw[0], cw[1], cw[2], cw[3]);
		}
	}
}

void Atlantis::Jettison ()
{
	switch (status) {
	case 0:
	case 3:               // nothing to do
		break;
	case 1:               // abandon boosters
		SeparateBoosters (oapiGetSimTime()-srb_ignition_time); 
		break;
	case 2:               // abandon tank
		SeparateTank();
		break;
	}
}

// Update moving parts of the orbiter's visual: payload bay doors and gear
// This should only be called when the visual exists, e.g. from within
// ovcVisualCreated or ovcAnimate

void Atlantis::UpdateMesh ()
{
	int grp_ofs;
	double delta;
	MESHGROUP_TRANSFORM mt;

	if (delta = door_proc-vdoor_proc) { // need to bring payload bay doors up to date
		grp_ofs = ORBITER_MESHGROUP_DOOR;
		float da = (float)(delta*DOOR_MAX_ANGLE);
		mt.nmesh = 0;
		mt.ngrp = grp_ofs+1; // left door
		mt.transform = MESHGROUP_TRANSFORM::ROTATE;
		mt.P.rotparam.axis = _V(0,0,1);
		mt.P.rotparam.angle = da;
		mt.P.rotparam.ref.x = -2.8;
		mt.P.rotparam.ref.y = 1.35;
		mt.P.rotparam.ref.z = 0;
		MeshgroupTransform (vis, mt);
		mt.ngrp = grp_ofs+4; // left panel 1
		mt.P.rotparam.angle = 0.9f*da;
		MeshgroupTransform (vis, mt);
		mt.ngrp = grp_ofs+5; // left panel 2
		mt.P.rotparam.angle = 0.95f*da;
		MeshgroupTransform (vis, mt);
		mt.ngrp = grp_ofs; // right door
		mt.P.rotparam.angle = -da;
		mt.P.rotparam.ref.x = 2.8;
		MeshgroupTransform (vis, mt);
		mt.ngrp = grp_ofs+2; // right panel 1
		mt.P.rotparam.angle = -0.9f*da;
		MeshgroupTransform (vis, mt);
		mt.ngrp = grp_ofs+3; // right panel 2
		mt.P.rotparam.angle = -0.95f*da;
		MeshgroupTransform (vis, mt);
		vdoor_proc = door_proc;
	}

	if (gear_proc-vgear_proc) { // need to bring landing gear up to date
		mt.nmesh = 0;
		mt.transform = MESHGROUP_TRANSFORM::ROTATE;
		grp_ofs = ORBITER_MESHGROUP_GEAR;

		// gear doors - note that the doors open during the first part of
		// the animation sequence (gear_proc < 0.5), while the gears are
		// lowered during the second part (gear_proc > 0.3)
		double alpha_tgt = (gear_proc  < 0.5 ?  gear_proc*1.1*PI : 0.55*PI);
		double alpha_now = (vgear_proc < 0.5 ? vgear_proc*1.1*PI : 0.55*PI);
		float da = (float)(alpha_tgt-alpha_now);
		if (da) {
			// left nose wheel door
			mt.P.rotparam.axis  = _V(0,0.195,0.981);
			mt.P.rotparam.angle = da;
			mt.P.rotparam.ref   = _V(-0.52,-2.19,17);
			for (mt.ngrp = grp_ofs; mt.ngrp < grp_ofs+2; mt.ngrp++)
				MeshgroupTransform (vis, mt);

			// right nose wheel door inside
			mt.P.rotparam.angle = -da;
			mt.P.rotparam.ref   = _V(0.52,-2.19,17);
			for (mt.ngrp = grp_ofs+2; mt.ngrp < grp_ofs+4; mt.ngrp++)
				MeshgroupTransform (vis, mt);

			// right gear door
			mt.P.rotparam.axis  = _V(0.0,0.0328,0.9994);
			mt.P.rotparam.angle = -da*0.95f;
			mt.P.rotparam.ref   = _V(4.8f,-2.47f,-4.69f);
			for (mt.ngrp = grp_ofs+6; mt.ngrp < grp_ofs+8; mt.ngrp++)
				MeshgroupTransform (vis, mt);

			// left gear door
			mt.P.rotparam.angle = da*0.95f;
			mt.P.rotparam.ref   = _V(-4.8f,-2.47f,-4.69f);
			for (mt.ngrp = grp_ofs+8; mt.ngrp < grp_ofs+10; mt.ngrp++)
				MeshgroupTransform (vis, mt);
		}

		// landing gear
		alpha_tgt = (gear_proc  > 0.3 ? ( gear_proc-0.3)*1.43*0.53*PI : 0.0);
		alpha_now = (vgear_proc > 0.3 ? (vgear_proc-0.3)*1.43*0.53*PI : 0.0);
		da = (float)(alpha_tgt-alpha_now);
		if (da) {
			// nose wheel
			mt.P.rotparam.angle = da;
			mt.P.rotparam.axis = _V(1,0,0);
			mt.P.rotparam.ref = _V(0,-1.5,16.95);
			for (mt.ngrp = grp_ofs+4; mt.ngrp < grp_ofs+6; mt.ngrp++)
				MeshgroupTransform (vis, mt);

			// left+right gear
			mt.P.rotparam.ref = _V(0,-2,-4.3);
			for (mt.ngrp = grp_ofs+10; mt.ngrp < grp_ofs+14; mt.ngrp++)
				MeshgroupTransform (vis, mt);
		}
		vgear_proc = gear_proc;
	}
}

void Atlantis::OperateLandingGear ()
{
	if (status < 3) return;
	// operate landing gear only once the orbiter is free from the tank

	if ((gear_status == DOOR_OPENING || gear_status == DOOR_CLOSING) && !gear_moving) {
		gear_moving = true;
		RegisterAnimation();
	}
}

void Atlantis::RevertLandingGear ()
{
	if (status < 3) return;
	// operate landing gear only once the orbiter is free from the tank

	gear_status = (gear_status == DOOR_CLOSED || gear_status == DOOR_CLOSING ?
				   DOOR_OPENING : DOOR_CLOSING);
	OperateLandingGear();
}

void Atlantis::OperateCargoDoors (void)
{
	if ((door_status == DOOR_OPENING || door_status == DOOR_CLOSING) && !door_moving) {
		door_moving = true;
		RegisterAnimation();
	}
}

void Atlantis::RevertCargoDoors (void)
{
	door_status = (door_status == DOOR_CLOSED || door_status == DOOR_CLOSING ?
		DOOR_OPENING : DOOR_CLOSING);
	 OperateCargoDoors();
}


// ==============================================================
// API interface
// ==============================================================

// Initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new Atlantis (hvessel, flightmodel);
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (Atlantis*)vessel;
}

// Set the capabilities of the vessel class
DLLCLBK void ovcSetClassCaps (VESSEL *vessel, FILEHANDLE cfg)
{
	//((Atlantis*)vessel)->SetLaunchConfiguration();
}

// Read status from scenario file
DLLCLBK void ovcLoadState (VESSEL *vessel, FILEHANDLE scn, VESSELSTATUS *def_vs)
{
	((Atlantis*)vessel)->LoadState (scn, def_vs);
}

// Save status to scenario file
DLLCLBK void ovcSaveState (VESSEL *vessel, FILEHANDLE scn)
{
	((Atlantis*)vessel)->SaveState (scn);
}

DLLCLBK void ovcTimestep (VESSEL *vessel, double simt)
{
	((Atlantis*)vessel)->Timestep (simt);
}

DLLCLBK void ovcAnimate (VESSEL *vessel, double simt)
{
	((Atlantis*)vessel)->UpdateMesh ();
}

DLLCLBK void ovcVisualCreated (VESSEL *vessel, VISHANDLE vis, int refcount)
{
	if (refcount > 1) return; // we don't allow more than one visual per object
	Atlantis *sts = (Atlantis*)vessel;
	sts->vis = vis;
	sts->vdoor_proc = 0.0;
	sts->vgear_proc = 0.0;
	sts->UpdateMesh ();
}

DLLCLBK void ovcVisualDestroyed (VESSEL *vessel, VISHANDLE vis, int refcount)
{
	Atlantis *sts = (Atlantis*)vessel;
	sts->vis = NULL;
}

// Keyboard interface handler (buffered key events)
DLLCLBK int ovcConsumeBufferedKey (VESSEL *vessel, DWORD key, bool down, char *kstate)
{
	if (!down) return 0; // only process keydown events

	Atlantis *sts = (Atlantis*)vessel;

	if (KEYMOD_SHIFT (kstate)) {
	} else if (KEYMOD_CONTROL (kstate)) {
		// insert ctrl key combinations here
	} else { // unmodified keys
		switch (key) {
		case OAPI_KEY_G:  // "Landing gear"
			sts->RevertLandingGear ();
			return 1;
		case OAPI_KEY_J:  // "Jettison"
			sts->bManualSeparate = true;
			return 1;
		case OAPI_KEY_K:  // "Cargo bay doors"
			sts->RevertCargoDoors ();
			return 1;
		}
	}
	return 0;
}
