// ==============================================================
//                 ORBITER MODULE: Atlantis
//                  Part of the ORBITER SDK
//            Copyright (C) 2001 Martin Schweiger
//                   All rights reserved
//
// Atlantis.h
// Class interface of Atlantis (Space Shuttle) vessel class
// module and associated subclasses (SRB, tank)
// ==============================================================

#ifndef __ATLANTIS_H
#define __ATLANTIS_H

#include "orbitersdk.h"
#include <math.h>

// ==========================================================
// Some Orbiter-related parameters
// ==========================================================

const double ORBITER_EMPTY_MASS = 104326.0;
// Orbiter empty mass [kg]

const double ORBITER_MAX_PROPELLANT_MASS = 20000.0;
// Amount of fuel the orbiter can hold in internal tanks
// I don't have this figure, so this is purely speculative ...

const double ORBITER_OMS_THRUST = 53400.0;
// Thrust rating for Orbital Maneuvering System [N]

const double ORBITER_RCS_THRUST = 7740.0;
// Thrust rating for attitude thrusters (Reaction Control System) [N]

const double ORBITER_OMS_ISP = 5000.0;
// Fuel-specific impulse for Orbital Maneuvering System [m/s]
// In fact this should be the value for monomethyl hydrazine/nitrogen tetroxide

const double GEAR_OPERATING_SPEED = 0.3;
// Opening/closing speed of landing gear (1/sec)
// => gear cycle ~ 3 sec

const double DOOR_OPERATING_SPEED = 0.016;
// Opening/closing speed of payload bay doors (1/sec)
// => door cycle = 63 sec

const double DOOR_MAX_ANGLE = RAD*160.0;
// Max payload bay door opening angle (rad)

const VECTOR3 ORBITER_CS = {234.8,389.1,68.2};
// Orbiter cross sections (projections into principal axes) [m^2]

const VECTOR3 ORBITER_CS_GEAR = {10.0,0.0,3.0};
// Contribution of fully extended landing gear to cross sections

const double ORBITER_CW[4] = {0.13, 0.5, 1.4, 1.4};
// Orbiter wind resistance coefficients in principal directions (+z, -z, +-x, +-y)

const double ORBITER_CW_GEAR[4] = {0.04, 0.04, 0.05, 0.0};
// Contribution of fully extended landing gear to wind resistance

// ==========================================================
// Some Tank-related parameters
// ==========================================================

const double TANK_EMPTY_MASS = 35425.0;
// Main tank empty mass

// ==========================================================
// Some SRB-related parameters
// ==========================================================

const double SRB_MAX_PROPELLANT_MASS = 502126.0;
// SRB propellant mass [kg]

const double SRB_EMPTY_MASS = 87543.0;
// SRB empty mass [kg]

const double SRB_ISP = 2812.77;
// SRB fuel-specific impulse

const double SRB_LIFTOFF_THRUST = 1202020.0*9.81;
// SRB thrust at liftoff (from Earth) [N]

const double SRB_FREE_THRUST = SRB_LIFTOFF_THRUST*1.25;
// SRB thrust in vacuum (guessed)

const double SRB_THRUST_PRESSURE_EXP = log(SRB_LIFTOFF_THRUST/SRB_FREE_THRUST)/101.4e3;
// used for calculating exponential relationship between
// atmospheric pressure and thrust

const double SRB_STABILISATION_TIME = 1.0;
// time between SRB ignition and full thrust

const double SRB_SEPARATION_TIME = 127.0;
// time between SRB ignition and separation

// ==========================================================
// Mesh offsets for various configurations
// ==========================================================

const VECTOR3 OFS_ZERO             = { 0.0, 0.0,  0.0  };
const VECTOR3 OFS_LAUNCH_ORBITER   = { 0.0, 6.22,-7.795};
const VECTOR3 OFS_LAUNCH_TANK      = { 0.0,-1.91, 5.72 };
const VECTOR3 OFS_LAUNCH_RIGHTSRB  = { 6.8,-1.91,-5.68 };
const VECTOR3 OFS_LAUNCH_LEFTSRB   = {-6.8,-1.91,-5.68 };
const VECTOR3 OFS_WITHTANK_ORBITER = { 0.0, 5.79,-9.185};
const VECTOR3 OFS_WITHTANK_TANK    = { 0.0,-2.34, 4.33 };
const VECTOR3 ORBITER_DOCKPOS      = { 0.0, 2.44,10.44 };

// ==========================================================
// Mesh group indices for some components
// ==========================================================

const int ORBITER_MESHGROUP_DOOR = 90;
const int ORBITER_MESHGROUP_GEAR = 96;

// ==========================================================
// Interface for derived vessel class: Atlantis
// ==========================================================

class Atlantis: public VESSEL {
public:
	Atlantis (OBJHANDLE hObj, int fmodel);
	void LoadState (FILEHANDLE scn, VESSELSTATUS *vs);
	void SaveState (FILEHANDLE scn);
	void SetLaunchConfiguration (void);
	void SetPostLaunchConfiguration (double srbtime);
	void SetOrbiterTankConfiguration (void);
	void SetOrbiterConfiguration (void);
	void AddOrbiterVisual (const VECTOR3 &ofs, bool maineng, bool atteng);
	void AddTankVisual (const VECTOR3 &ofs);
	void AddSRBVisual (const VECTOR3 &ofs);
	void SeparateBoosters (double srb_time);
	void SeparateTank (void);
	void Timestep (double simt);
	void Jettison ();
	void UpdateMesh ();
	void OperateLandingGear ();
	void RevertLandingGear ();
	void OperateCargoDoors ();
	void RevertCargoDoors ();
	int status; // 0=launch configuration
	            // 1=SRB's engaged
	            // 2=SRB's separated
	            // 3=Tank separated (orbiter only)
	bool bManualSeparate; // flag for user-induced booster or tank separation

	double *srb_thrust; // pointer to SRB thrust component
	double srb_ignition_time; // simulation time at which SRB's were ignited
	WORD srb_id1, srb_id2;

	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } door_status, gear_status;
	double door_proc, vdoor_proc; // logical and visual payload bay door opening state (0=closed, 1=open)
	bool door_moving;   // flag for door moving
	bool door_endanim;  // flag for animation stop
	double gear_proc, vgear_proc;
	bool gear_moving, gear_endanim;
	VISHANDLE vis;      // handle for visual - note: we assume that only one visual per object is created!
	MESHHANDLE hOrbiterMesh, hTankMesh, hSRBMesh; // mesh handles
};

// ==========================================================
// Interface for derived vessel class: Atlantis_SRB
// ==========================================================

class Atlantis_SRB: public VESSEL {
public:
	Atlantis_SRB (OBJHANDLE hObj);
	// Construct interface from existing object

	void SetClassSpecs (void);
	void Timestep (double simt);
	void SetRefTime (void);

private:
	double ignition_time;       // simulation time at which SRB was ignited
	double srb_separation_time; // simulation time at which SRB separation was initiated
	bool bMainEngine;           // main engine firing?
	bool bSeparationEngine;     // separation thrusters firing?
	WORD sep_id1, sep_id2;
	MESHHANDLE hSRBMesh;
};

// ==========================================================
// Interface for derived vessel class: Atlantis_Tank
// ==========================================================

class Atlantis_Tank: public VESSEL {
public:
	Atlantis_Tank (OBJHANDLE hObj);
	// Construct interface from existing object

	void SetClassSpecs (void);
	void Timestep (double simt);

private:
	MESHHANDLE hTankMesh;
};

#endif // !__ATLANTIS_H
