// ==============================================================
//                 ORBITER MODULE: Atlantis
//                  Part of the ORBITER SDK
//            Copyright (C) 2001 Martin Schweiger
//                   All rights reserved
//
// Atlantis_SRB.cpp
// Reference implementation of Atlantis SRB(Space Shuttle - Solid
// Rocket Booster) vessel class module
// Note: This module takes control of the SRB after separation
// from the Shuttle's main tank.
// ==============================================================

#include "Atlantis.h"
#include "math.h"
#include "stdio.h"

// ==============================================================
// Specialised vessel class Atlantis_SRB
// ==============================================================

// Constructor
Atlantis_SRB::Atlantis_SRB (OBJHANDLE hObj)
: VESSEL(hObj)
{
	// preload mesh
	hSRBMesh = oapiLoadMeshGlobal ("Atlantis_srb");
}

// Set SRB class specs
void Atlantis_SRB::SetClassSpecs (void)
{
	SetSize (23.0);
	SetEmptyMass (SRB_EMPTY_MASS);

	SetMaxFuelMass (SRB_MAX_PROPELLANT_MASS);
	// Note that the SRB instance is only created after separation from
	// the main assembly, so the actual fuel mass will always be
	// much smaller

	SetISP (SRB_ISP);

	SetMaxThrust (ENGINE_MAIN, SRB_LIFTOFF_THRUST);
	// This is actually the thrust at liftoff.
	// Note that the actual thrust will be smaller since
	// 1. Thrust reduces by 1/3 at 50 sec after liftoff
	// 2. After separation, thrust will start to drop as
	//    propellant is exhausted

	SetMaxThrust (ENGINE_RETRO, 0);
	SetMaxThrust (ENGINE_HOVER, 0);
	SetMaxThrust (ENGINE_ATTITUDE, 1e4);
	SetCOG_elev (21.1);
	SetCW (0.1, 0.3, 1.4, 1.4);
	VECTOR3 cs = {162.1,162.1,26.6};
	SetCrossSections (cs);
	VECTOR3 rd = {0.7,0.7,0.1};
	SetRotDrag (rd);
	VECTOR3 pmi = {154.3,154.3,1.83};
	SetPMI (pmi);
	SetPitchMomentScale (1e-4);
	SetBankMomentScale (1e-4);

	VECTOR3 co = {0,0,0};
	SetCameraOffset (co);
	// Note that the camera offset should not be required
	// since the SRB doesn't define a 'cockpit'

	VECTOR3 touchdown_p1 = {0,-2,-21.1};
	VECTOR3 touchdown_p2 = {-2,1,-21.1};
	VECTOR3 touchdown_p3 = { 2,1,-21.1};
	SetTouchdownPoints (touchdown_p1, touchdown_p2, touchdown_p3);

	SetLiftCoeffFunc (0);

	AddMesh (hSRBMesh);
	VECTOR3 ofs = {0,0,-21};
	AddExhaustRef (EXHAUST_MAIN, ofs, 16, 2);
	bMainEngine = true;

	VECTOR3 dir = {-1,0,0};
	ofs.x = -2.1, ofs.y = 0, ofs.z = -10;
	sep_id1 = AddExhaustRef (EXHAUST_CUSTOM, ofs, 2, 0.3, &dir);
	ofs.z = 10;
	sep_id2 = AddExhaustRef (EXHAUST_CUSTOM, ofs, 2, 0.3, &dir);
	bSeparationEngine = true;
	srb_separation_time = oapiGetSimTime();
}

void Atlantis_SRB::Timestep (double simt)
{
	extern void GetSRB_State (double, double&, double&);
	
	if (bMainEngine) {
		double srb_time = simt-ignition_time;
		if (srb_time >= 136.0) {
			SetEngineLevel (ENGINE_MAIN, 0);
			bMainEngine = false;
			// After the propellant is burnt out we should be airborne.
			// Now we can prepare touchdown points for "landing"
			SetCOG_elev (-5.0);
			SetTouchdownPoints (_V(0,9,3), _V(-1,1,-3), _V(1,1,-3));
		} else {
			double thrust_level, prop_level;
			GetSRB_State (srb_time, thrust_level, prop_level);
			SetEngineLevel (ENGINE_MAIN, thrust_level);
		}
		if (bSeparationEngine) {
			double srb_dt = simt - srb_separation_time;
			if (srb_dt > 2.0) {
				DelExhaustRef (EXHAUST_CUSTOM, sep_id1);
				DelExhaustRef (EXHAUST_CUSTOM, sep_id2);
				bSeparationEngine = false;
			} else {
				double fac = sqrt (1.0-srb_dt/2.0);
				SetExhaustScales (EXHAUST_CUSTOM, sep_id1, 2*fac, 0.3*fac);
				SetExhaustScales (EXHAUST_CUSTOM, sep_id2, 2*fac, 0.3*fac);
			}
		}
	}
	if (GetAltitude() < 0.0) oapiDeleteVessel (GetHandle());
}

// reconstruct ignition time from fuel level
void Atlantis_SRB::SetRefTime (void)
{
	extern int SRB_nt;
	extern double SRB_Seq[], SRB_Prop[], SRB_PrpSCL[];

	double fuel = GetFuelMass()/GetMaxFuelMass();
	for (int i = 1; i < SRB_nt; i++)
		if (fuel >= SRB_Prop[i]) break;
	double burn_time = SRB_Seq[i] + (fuel-SRB_Prop[i])/SRB_PrpSCL[i-1];
	ignition_time = oapiGetSimTime()-burn_time;
}

// ==============================================================
// API interface
// ==============================================================

// Initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new Atlantis_SRB (hvessel);
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (Atlantis_SRB*)vessel;
}

// Vessel class caps
DLLCLBK void ovcSetClassCaps (VESSEL *vessel, FILEHANDLE cfg)
{
	((Atlantis_SRB*)vessel)->SetClassSpecs ();
}

// Vessel status initialisation
DLLCLBK void ovcSetState (VESSEL *vessel, const VESSELSTATUS *status)
{
	vessel->DefSetState (status);	        // default initialisation
	((Atlantis_SRB*)vessel)->SetRefTime ();	// reconstruct ignition time from fuel level
}

DLLCLBK void ovcTimestep (VESSEL *vessel, double simt)
{
	((Atlantis_SRB*)vessel)->Timestep (simt);
}