// ==============================================================
//                ORBITER MODULE: DeltaGlider
//                  Part of the ORBITER SDK
//          Copyright (C) 2001-2002 Martin Schweiger
//                   All rights reserved
//
// DeltaGlider.cpp
// Reference implementation of delta glider vessel class module
// ==============================================================

#define STRICT

#include "DeltaGlider.h"
#include <stdio.h>
#include <math.h>

#define LOADBMP(id) (LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (id)))

// ==============================================================
// Global parameters
// ==============================================================

GDIParams g_Param;

// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,   -0.4,     0,   0.8,     1,   0.8,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	for (int i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
	return CL[i] + (aoa-AOA[i])*SCL[i];
}

// ==============================================================
// Specialised vessel class DeltaGlider
// ==============================================================

// Constructor
DeltaGlider::DeltaGlider (OBJHANDLE hObj, int fmodel)
: VESSEL (hObj, fmodel)
{
	aoa_ind = PI;
	slip_ind = PI*0.5;
	slope = slope_ind = PI;
	navmode_bt_status = 0xffff;
	gear_status     = DOOR_CLOSED;
	gear_proc       = 0.0;
	nose_status     = DOOR_CLOSED;
	nose_proc       = 0.0;
	lock_status     = DOOR_CLOSED;
	lock_proc       = 0.0;
	DefineAnimations();
	for (int i = 0; i < nsurf; i++) srf[i] = 0;
}

// Destructor
DeltaGlider::~DeltaGlider ()
{
	ReleaseSurfaces();
}

// Read status from scenario file
void DeltaGlider::LoadState (FILEHANDLE scn, VESSELSTATUS *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
        if (!strnicmp (line, "NOSECONE", 8)) {
			sscanf (line+8, "%d%lf", &nose_status, &nose_proc);
		} else if (!strnicmp (line, "GEAR", 4)) {
			sscanf (line+4, "%d%lf", &gear_status, &gear_proc);
		} else if (!strnicmp (line, "AIRLOCK", 7)) {
			sscanf (line+7, "%d%lf", &lock_status, &lock_proc);
        } else {
            ParseScenarioLine (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
    }

	// update animation states
	SetAnimState (anim_gear, gear_proc);
	SetAnimState (anim_nose, nose_proc);
	SetAnimState (anim_lock, lock_proc);
}

// Write status to scenario file
void DeltaGlider::SaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	SaveDefaultState (scn);

	// custom parameters
	sprintf (cbuf, "%d %0.4f", nose_status, nose_proc);
	oapiWriteScenario_string (scn, "NOSECONE", cbuf);

	sprintf (cbuf, "%d %0.4f", gear_status, gear_proc);
	oapiWriteScenario_string (scn, "GEAR", cbuf);

	sprintf (cbuf, "%d %0.4f", lock_status, lock_proc);
	oapiWriteScenario_string (scn, "AIRLOCK", cbuf);
}

// Set vessel class parameters
void DeltaGlider::SetClassCaps (FILEHANDLE cfg)
{
	SetSize (9.0);
	SetEmptyMass (8000.0);
	SetMaxFuelMass (10000.0);
	SetISP (5e4);
	SetMaxThrust (ENGINE_MAIN, MAX_MAIN_THRUST);
	SetMaxThrust (ENGINE_RETRO, MAX_RETRO_THRUST);
	SetMaxThrust (ENGINE_HOVER, MAX_HOVER_THRUST);
	SetMaxThrust (ENGINE_ATTITUDE, MAX_ATTITUDE_THRUST);
	SetCOG_elev (2.47);
	SetCW (0.09, 0.09, 2, 1.4);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(53.0,186.9,25.9));
	SetRotDrag (_V(1.5,1.5,1.5));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (5e-5);
		SetBankMomentScale (5e-5);
	}
	SetPMI (_V(15.5,22.1,7.7));
	SetTrimScale (0.05);
	SetCameraOffset (_V(0,1.5,6.8));
	SetLiftCoeffFunc (LiftCoeff);
	SetDockParams (_V(0,-0.39,9.84), _V(0,0,1), _V(0,1,0));
	SetTouchdownPoints (_V(0,-2.47,9), _V(-4,-2.47,-5), _V(4,-2.47,-5));
	AddMesh (oapiLoadMeshGlobal ("DeltaGlider"));

	AddExhaustRef (EXHAUST_MAIN,  _V(-1, 0.2,-7.7), 12, 1);
	AddExhaustRef (EXHAUST_MAIN,  _V( 1, 0.2,-7.7), 12, 1);
	AddExhaustRef (EXHAUST_RETRO, _V(-3,-0.2, 5.6), 3, 0.4);
	AddExhaustRef (EXHAUST_RETRO, _V( 3,-0.2, 5.6), 3, 0.4);
	AddExhaustRef (EXHAUST_HOVER, _V( 0,-1.5, 3), 6, 0.5);
	AddExhaustRef (EXHAUST_HOVER, _V(-3,-1.2,-3), 6, 0.5);
	AddExhaustRef (EXHAUST_HOVER, _V( 3,-1.2,-3), 6, 0.5);

	UINT idx;
	idx = AddAttExhaustRef (_V(-0.75,-0.2,9.65), _V(0,1,0), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
	AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
	idx = AddAttExhaustRef (_V( 0.75,-0.2,9.65), _V(0,1,0), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
	AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
	idx = AddAttExhaustRef (_V(-0.75,-0.6,9.65), _V(0,-1,0), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
	AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
	idx = AddAttExhaustRef (_V( 0.75,-0.6,9.65), _V(0,-1,0), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
	AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
	idx = AddAttExhaustRef (_V(-0.1,0.65,-7.3), _V(0,1,0), 0.3, 0.3);
	AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
	AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
	idx = AddAttExhaustRef (_V( 0.1,0.65,-7.3), _V(0,1,0), 0.3, 0.3);
	AddAttExhaustMode (idx, ATTMODE_ROT, 0, 0); // pitch up
	AddAttExhaustMode (idx, ATTMODE_LIN, 1, 1); // move down
	idx = AddAttExhaustRef (_V(-0.1,-0.45,-7.3), _V(0,-1,0), 0.3, 0.3);
	AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
	AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
	idx = AddAttExhaustRef (_V( 0.1,-0.45,-7.3), _V(0,-1,0), 0.3, 0.3);
	AddAttExhaustMode (idx, ATTMODE_ROT, 0, 1); // pitch down
	AddAttExhaustMode (idx, ATTMODE_LIN, 1, 0); // move up
	idx = AddAttExhaustRef (_V(-5.1,0.3,0.4), _V(0,1,0), 0.3, 0.3);
	AddAttExhaustMode (idx, ATTMODE_ROT, 2, 1); // roll left
	idx = AddAttExhaustRef (_V( 5.1,0.3,0.4), _V(0,1,0), 0.3, 0.3);
	AddAttExhaustMode (idx, ATTMODE_ROT, 2, 0); // roll right
	idx = AddAttExhaustRef (_V( 5.1,-0.7,0.4), _V(0,-1,0), 0.3, 0.3);
	AddAttExhaustMode (idx, ATTMODE_ROT, 2, 1); // roll left
	idx = AddAttExhaustRef (_V(-5.1,-0.7,0.4), _V(0,-1,0), 0.3, 0.3);
	AddAttExhaustMode (idx, ATTMODE_ROT, 2, 0); // roll right
	idx = AddAttExhaustRef (_V(-0.81,-0.4,9.65), _V(-1,0,0), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_ROT, 1, 1); // yaw right
	AddAttExhaustMode (idx, ATTMODE_LIN, 0, 0); // move right
	idx = AddAttExhaustRef (_V( 0.81,-0.4,9.65), _V( 1,0,0), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_ROT, 1, 0); // yaw left
	AddAttExhaustMode (idx, ATTMODE_LIN, 0, 1); // move left
	idx = AddAttExhaustRef (_V(-2.1,0.4,-5.6), _V(-1,0,0), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_ROT, 1, 0); // yaw left
	AddAttExhaustMode (idx, ATTMODE_LIN, 0, 0); // move right
	idx = AddAttExhaustRef (_V( 2.1,0.4,-5.6), _V( 1,0,0), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_ROT, 1, 1); // yaw right
	AddAttExhaustMode (idx, ATTMODE_LIN, 0, 1); // move left
	idx = AddAttExhaustRef (_V(-0.77,-0.4,9.8), _V(0,0,1), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_LIN, 2, 1); // move back
	idx = AddAttExhaustRef (_V( 0.77,-0.4,9.8), _V(0,0,1), 0.2, 0.2);
	AddAttExhaustMode (idx, ATTMODE_LIN, 2, 1); // move back
	idx = AddAttExhaustRef (_V(0,-0.1,-7.6), _V(0,0,-1), 0.2,0.2);
	AddAttExhaustMode (idx, ATTMODE_LIN, 2, 0); // move forward
	idx = AddAttExhaustRef (_V(0,0.32,-7.6), _V(0,0,-1), 0.2,0.2);
	AddAttExhaustMode (idx, ATTMODE_LIN, 2, 0); // move forward
}

// Define animation sequences for moving parts
void DeltaGlider::DefineAnimations ()
{
	static UINT NosewheelGrp1[4] = {16,35,61,10};
	static ANIMCOMP nosewheel1 = { // nose wheel and front flap
		NosewheelGrp1, 4, 0.3, 0.9,
		0,-1.00,8.553,              // rotation reference
		1,0,0,                      // rotation axis
		(float)(-0.553*PI),         // rotation range
		0,                          // mesh no.
		0,                          // not used
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT NosewheelGrp2[2] = {13,34};
	static ANIMCOMP nosewheel2 = { // nose wheel left side flap
		NosewheelGrp2, 2, 0.0, 0.4,
		-0.3, -1.12046, 7.02812,    // rotation reference
		0,0.045953,0.99894,         // rotation axis
		(float)(0.5*PI),            // rotation range
		0,                          // mesh no.
		0,                          // not used
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT NosewheelGrp3[2] = {12,33};
	static ANIMCOMP nosewheel3 = { // nose wheel right side flap
		NosewheelGrp3, 2, 0.0, 0.42,
		0.3, -1.12046, 7.02812,     // rotation reference
		0,0.045953,0.99894,         // rotation axis
		(float)(-0.5*PI),           // rotation range
		0,                          // mesh no.
		0,                          // not used
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT RightwheelGrp1[4] = {8,18,39,59};
	static ANIMCOMP rightwheel1 = { // right wheel and flap
		RightwheelGrp1, 4, 0.2, 1.0,
		3.631,-0.699,-4.9499,
		0,0,1,
		(float)(0.5*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT RightwheelGrp2[2] = {14,38};
	static ANIMCOMP rightwheel2 = { // right wheel inner flap
		RightwheelGrp2, 2, 0.04, 0.5,
		1.82,-0.76,-4.4621,
		0,0,1,
		(float)(-0.5*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT LeftwheelGrp1[4] = {9,17,36,60};
	static ANIMCOMP leftwheel1 = { // left wheel and flap
		LeftwheelGrp1, 4, 0.17, 0.96,
		-3.631,-0.699,-4.9499,
		0,0,1,
		(float)(-0.5*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT LeftwheelGrp2[2] = {15,37};
	static ANIMCOMP leftwheel2 = { // left wheel inner flap
		LeftwheelGrp2, 2, 0.02, 0.48,
		-1.82,-0.76,-4.4621,
		0,0,1,
		(float)(0.5*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE // transform type
	};
	static UINT NoseconeGrp1[2] = {11,40};
	static ANIMCOMP nosecone1 = { // upper right nosecone segment
		NoseconeGrp1, 2, 0.0, 1.0,
		0.601,-0.141,9.837,
		-0.351,0.351,0,
		(float)(-0.80*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT NoseconeGrp2[2] = {19,43};
	static ANIMCOMP nosecone2 = { // lower right nosecone segment
		NoseconeGrp2, 2, 0.0, 1.0,
		0.601,-0.638,9.837,
		0.351,0.351,0,
		(float)(-0.80*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT NoseconeGrp3[2] = {20,42};
	static ANIMCOMP nosecone3 = { // lower left nosecone segment
		NoseconeGrp3, 2, 0.0, 1.0,
		-0.601,-0.638,9.837,
		0.351,-0.351,0,
		(float)(-0.80*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT NoseconeGrp4[2] = {21,41};
	static ANIMCOMP nosecone4 = { // upper left nosecone segment
		NoseconeGrp4, 2, 0.0, 1.0,
		-0.601,-0.141,9.837,
		-0.351,-0.351,0,
		(float)(-0.80*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT AirlockGrp1[1] = {7};
	static ANIMCOMP airlock = { // outer airlock
		AirlockGrp1, 1, 0.0, 1.0,
		//-0.371,-0.018,9.85,
		-0.169,0.020,9.849,
		1,0,0,
		(float)(0.45*PI),
		0,
		0,
		MESHGROUP_TRANSFORM::ROTATE
	};

	// Register animation for landing gear
	anim_gear = RegisterAnimSequence (1.0); // gear is fully extended in the mesh
	AddAnimComp (anim_gear, &nosewheel1);   // nose wheel component 1
	AddAnimComp (anim_gear, &nosewheel2);   // nose wheel component 2
	AddAnimComp (anim_gear, &nosewheel3);   // nose wheel component 3
	AddAnimComp (anim_gear, &rightwheel1);  // right wheel component 1
	AddAnimComp (anim_gear, &rightwheel2);  // right wheel component 2
	AddAnimComp (anim_gear, &leftwheel1);   // left wheel component 1
	AddAnimComp (anim_gear, &leftwheel2);   // left wheel component 2

	anim_nose = RegisterAnimSequence (0.0); // nose cone is closed in mesh
	AddAnimComp (anim_nose, &nosecone1);    // nose cone component 1
	AddAnimComp (anim_nose, &nosecone2);    // nose cone component 2
	AddAnimComp (anim_nose, &nosecone3);    // nose cone component 3
	AddAnimComp (anim_nose, &nosecone4);    // nose cone component 4

	anim_lock = RegisterAnimSequence (0.0); // airlock is closed in mesh
	AddAnimComp (anim_lock, &airlock);      // airlock component 1
}

void DeltaGlider::ReleaseSurfaces ()
{
	for (int i = 0; i < nsurf; i++)
		if (srf[i]) {
			oapiDestroySurface (srf[i]);
			srf[i] = 0;
		}
}

void DeltaGlider::Timestep (double simt)
{
	VECTOR3 v;
	double hspd, vspd;
	if (GetHorizonAirspeedVector (v)) {
		hspd = hypot (v.x, v.z);
		vspd = v.y;
		slope = (hspd ? atan (vspd/hspd) : 0.0);
	} else slope = 0.0;
	if ((alt = GetAltitude()) > 0.0 && slope < 0.0)
		range = -alt*hspd/vspd;
	else range = -1.0;

	// animate landing gear
	if (gear_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * GEAR_OPERATING_SPEED;
		if (gear_status == DOOR_CLOSING) {
			if (gear_proc > 0.0)
				gear_proc = max (0.0, gear_proc-da);
			else {
				gear_status = DOOR_CLOSED;
				oapiTriggerPanelRedrawArea (1, AID_GEARINDICATOR);
			}
		} else  { // door opening
			if (gear_proc < 1.0)
				gear_proc = min (1.0, gear_proc+da);
			else {
				gear_status = DOOR_OPEN;
				oapiTriggerPanelRedrawArea (1, AID_GEARINDICATOR);
			}
		}
		SetAnimState (anim_gear, gear_proc);
	}

	// animate nose cone
	if (nose_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * NOSE_OPERATING_SPEED;
		if (nose_status == DOOR_CLOSING) {
			if (nose_proc > 0.0)
				nose_proc = max (0.0, nose_proc-da);
			else {
				nose_status = DOOR_CLOSED;
				oapiTriggerPanelRedrawArea (1, AID_DOCKINDICATOR);
			}
		} else { // door opening
			if (nose_proc < 1.0)
				nose_proc = min (1.0, nose_proc+da);
			else {
				nose_status = DOOR_OPEN;
				oapiTriggerPanelRedrawArea (1, AID_DOCKINDICATOR);
			}
		}
		SetAnimState (anim_nose, nose_proc);
	}

	// animate airlock
	if (lock_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * AIRLOCK_OPERATING_SPEED;
		if (lock_status == DOOR_CLOSING) {
			if (lock_proc > 0.0)
				lock_proc = max (0.0, lock_proc-da);
			else {
				lock_status = DOOR_CLOSED;
				//oapiTriggerPanelRedrawArea (1, AID_AIRLOCKINDICATOR);
			}
		} else { // door opening
			if (lock_proc < 1.0)
				lock_proc = min (1.0, lock_proc+da);
			else {
				lock_status = DOOR_OPEN;
				//oapiTriggerPanelRedrawArea (1, AID_AIRLOCKINDICATOR);
			}
		}
		SetAnimState (anim_lock, lock_proc);
	}
}

void DeltaGlider::ActivateLandingGear (DoorStatus action)
{
	gear_status = action;
	oapiTriggerPanelRedrawArea (1, AID_GEARBUTTON);
	oapiTriggerPanelRedrawArea (1, AID_GEARINDICATOR);
}

void DeltaGlider::RevertLandingGear ()
{
	ActivateLandingGear (gear_status == DOOR_CLOSED || gear_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}

void DeltaGlider::ActivateDockingPort (DoorStatus action)
{
	nose_status = action;
	oapiTriggerPanelRedrawArea (1, AID_DOCKBUTTON);
	oapiTriggerPanelRedrawArea (1, AID_DOCKINDICATOR);
}

void DeltaGlider::RevertDockingPort ()
{
	ActivateDockingPort (nose_status == DOOR_CLOSED || nose_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}

void DeltaGlider::ActivateAirlock (DoorStatus action)
{
	lock_status = action;
	oapiTriggerPanelRedrawArea (1, AID_OUTERAIRLOCKBUTTON);
	//oapiTriggerPanelRedrawArea (1, AID_AIRLOCKINDICATOR);
}

void DeltaGlider::RevertAirlock ()
{
	ActivateAirlock (lock_status == DOOR_CLOSED || lock_status == DOOR_CLOSING ?
		             DOOR_OPENING : DOOR_CLOSING);
}

bool DeltaGlider::DecAttMode ()
{
	int mode = GetAttitudeMode();
	if (mode) {
		SetAttitudeMode (mode-1);
		return true;
	} else return false;
}

bool DeltaGlider::IncAttMode ()
{
	int mode = GetAttitudeMode();
	if (mode < 2) {
		SetAttitudeMode (mode+1);
		return true;
	} else return false;
}

void DeltaGlider::RedrawPanel_PropellantStatus (SURFHANDLE surf)
{
	char cbuf[16];
	int i;
	double m0   = GetMaxFuelMass();
	double m    = GetFuelMass();
	double lvl  = m/m0;
	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.font[0]);
	if (i = (int)(101.0*lvl)) {
		SelectObject (hDC, g_Param.brush[0]);
		SelectObject (hDC, GetStockObject (NULL_PEN));
		Rectangle (hDC, 0, 0, i+1, 8);
	}
	SetTextColor (hDC, RGB(0, 255, 0));
	SetTextAlign (hDC, TA_RIGHT);
	SetBkMode (hDC, TRANSPARENT);
	sprintf (cbuf, "%0.1f", 100.0*lvl);
	TextOut (hDC, 75, 11, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.0f", m);
	TextOut (hDC, 75, 30, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.2f", GetFuelRate());
	TextOut (hDC, 75, 49, cbuf, strlen(cbuf));
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_EngineStatus (SURFHANDLE surf)
{
	char cbuf[16];
	double lvlm, lvlr = 0.0, lvlh, FM, FH;
	double m = GetMass();
	int scl;
	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.font[0]);
	SelectObject (hDC, g_Param.brush[1]);
	SelectObject (hDC, GetStockObject (NULL_PEN));

	if (lvlm = GetEngineLevel (ENGINE_MAIN)) {
		scl = (int)(lvlm*82.0);
		Rectangle (hDC, 27, 82, 37, 82-scl);
	} else if (lvlr = GetEngineLevel (ENGINE_RETRO)) {
		scl = (int)(lvlr*22.0);
		Rectangle (hDC, 38, 80, 48, 80+scl);
	}
	if (lvlh = GetEngineLevel (ENGINE_HOVER)) {
		scl = (int)(lvlh*102.0);
		Rectangle (hDC, 76, 102, 86, 102-scl);
	}
	FM = lvlm ? lvlm*MAX_MAIN_THRUST : -lvlr*MAX_RETRO_THRUST;
	FH = lvlh*MAX_HOVER_THRUST;
	SetTextColor (hDC, RGB(0,255,0));
	SetTextAlign (hDC, TA_RIGHT);
	SetBkMode (hDC, TRANSPARENT);
	sprintf (cbuf, "%0.1f", (lvlm ? lvlm : -lvlr)*100.0);
	TextOut (hDC, 39, 108, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.1f", lvlh*100.0);
	TextOut (hDC, 106, 108, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.1f", FM*1e-3);
	TextOut (hDC, 39, 129, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.1f", FH*1e-3);
	TextOut (hDC, 106, 129, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.2f", FM/m);
	TextOut (hDC, 39, 150, cbuf, strlen(cbuf));
	sprintf (cbuf, "%0.2f", FH/m);
	TextOut (hDC, 106, 150, cbuf, strlen(cbuf));
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_AOA (SURFHANDLE surf)
{
	static const double dial_max = RAD*165.0;
	double dial_angle = 7.7 * GetAOA();
	if (fabs (dial_angle) > dial_max)
		dial_angle = (dial_angle > 0.0 ? dial_max:-dial_max);
	HDC hDC = oapiGetDC (surf);
	DrawNeedle (hDC, 28, 28, 26.0, PI-dial_angle, &aoa_ind);
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_Slip (SURFHANDLE surf)
{
	static const double dial_max = RAD*165.0;
	double dial_angle = 7.7 * GetSlipAngle();
	if (fabs (dial_angle) > dial_max)
		dial_angle = (dial_angle > 0.0 ? dial_max:-dial_max);
	HDC hDC = oapiGetDC (surf);
	DrawNeedle (hDC, 28, 28, 26.0, PI*0.5-dial_angle, &slip_ind);
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_Slope (SURFHANDLE surf)
{
	const double smax = 7.0*RAD, smin = -22.0*RAD;
	double alpha;
	if      (slope > smax) alpha = smax;
	else if (slope < smin) alpha = smin;
	else                   alpha = slope;
	HDC hDC = oapiGetDC (surf);
	DrawNeedle (hDC, 140, 19, 140.0, PI-alpha, &slope_ind, 0.2);
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_Range (SURFHANDLE surf)
{
	char cbuf[16];
	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.font[0]);
	SetTextColor (hDC, RGB(255, 0, 0));
	SetTextAlign (hDC, TA_RIGHT);
	SetBkMode (hDC, TRANSPARENT);
	sprintf (cbuf, "%+0.1f", DEG*slope);
	TextOut (hDC, 34, 0, cbuf, strlen(cbuf));
	if (range > 0.0 && range < 1e7) {
		sprintf (cbuf, "%0.0f", 0.001*range);
		TextOut (hDC, 34, 30, cbuf, strlen(cbuf));
	} else
		TextOut (hDC, 34, 30, "N/A", 3);
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::RedrawPanel_Navmode (SURFHANDLE surf)
{
	for (DWORD i = NAVMODE_KILLROT; i <= NAVMODE_HOLDALT; i++)
		if (GetNavmodeState (i))
				oapiBlt (surf, srf[1], ((i-1)/2) * 39, ((i-1)%2) * 39, (i-1)*37, 0, 37, 37);
}

void DeltaGlider::RedrawPanel_MFDButton (SURFHANDLE surf, int mfd, int side)
{
	HDC hDC = oapiGetDC (surf);
	SelectObject (hDC, g_Param.font[1]);
	SetTextColor (hDC, RGB(196, 196, 196));
	SetTextAlign (hDC, TA_CENTER);
	SetBkMode (hDC, TRANSPARENT);
	const char *label;
	for (int bt = 0; bt < 6; bt++) {
		if (label = oapiMFDButtonLabel (mfd, bt+side*6))
			TextOut (hDC, 13, 3+41*bt, label, strlen(label));
		else break;
	}
	oapiReleaseDC (surf, hDC);
}

void DeltaGlider::DrawNeedle (HDC hDC, int x, int y, double rad, double angle, double *pangle, double vdial)
{
	if (pangle) { // needle response delay
		double dt = oapiGetSimStep();
		if (fabs (angle - *pangle)/dt >= vdial)
			angle = (angle > *pangle ? *pangle+vdial*dt : *pangle-vdial*dt);
		*pangle = angle;
	}
	double dx = rad * cos(angle), dy = rad * sin(angle);
	SelectObject (hDC, g_Param.pen[1]);
	MoveToEx (hDC, x, y, 0); LineTo (hDC, x + (int)(0.85*dx+0.5), y - (int)(0.85*dy+0.5));
	SelectObject (hDC, g_Param.pen[0]);
	MoveToEx (hDC, x, y, 0); LineTo (hDC, x + (int)(dx+0.5), y - (int)(dy+0.5));
}

// ==============================================================
// DLL entry point
// ==============================================================

BOOL WINAPI DllMain (HINSTANCE hModule,
					 DWORD ul_reason_for_call,
					 LPVOID lpReserved)
{
	int i;

	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		g_Param.hDLL = hModule;
		// allocate GDI resources
		g_Param.col[2] = oapiGetColour(80,80,224);
		g_Param.col[3] = oapiGetColour(160,120,64);
		g_Param.font[0]  = CreateFont (-13, 0, 0, 0, 700, 0, 0, 0, 0, 0, 0, 0, 0, "Arial");
		g_Param.font[1]  = CreateFont (-10, 0, 0, 0, 400, 0, 0, 0, 0, 0, 0, 0, 0, "Arial");
		g_Param.brush[0] = CreateSolidBrush (RGB(0,255,0));    // green
		g_Param.brush[1] = CreateSolidBrush (RGB(255,0,0));    // red
		g_Param.brush[2] = CreateSolidBrush (RGB(80,80,224));  // blue
		g_Param.brush[3] = CreateSolidBrush (RGB(160,120,64));  // brown
		g_Param.pen[0] = CreatePen (PS_SOLID, 1, RGB(224,224,224));
		g_Param.pen[1] = CreatePen (PS_SOLID, 3, RGB(164,164,164));
		break;
	case DLL_PROCESS_DETACH:
		// deallocate GDI resources
		for (i = 0; i < 2; i++) DeleteObject (g_Param.font[i]);
		for (i = 0; i < 4; i++) DeleteObject (g_Param.brush[i]);
		for (i = 0; i < 2; i++) DeleteObject (g_Param.pen[i]);
		break;
	}
	return TRUE;
}

// ==============================================================
// API interface
// ==============================================================

// Initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new DeltaGlider (hvessel, flightmodel);
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (DeltaGlider*)vessel;
}

// Set the capabilities of the vessel class
DLLCLBK void ovcSetClassCaps (VESSEL *vessel, FILEHANDLE cfg)
{
	((DeltaGlider*)vessel)->SetClassCaps (cfg);
}

// Read status from scenario file
DLLCLBK void ovcLoadState (VESSEL *vessel, FILEHANDLE scn, VESSELSTATUS *def_vs)
{
	((DeltaGlider*)vessel)->LoadState (scn, def_vs);
}

// Save status to scenario file
DLLCLBK void ovcSaveState (VESSEL *vessel, FILEHANDLE scn)
{
	((DeltaGlider*)vessel)->SaveState (scn);
}

DLLCLBK void ovcTimestep (VESSEL *vessel, double simt)
{
	((DeltaGlider*)vessel)->Timestep (simt);
}

DLLCLBK void ovcNavmode (VESSEL *vessel, int mode, bool active)
{
	oapiTriggerPanelRedrawArea (0, AID_NAVMODE);
}

DLLCLBK void ovcHUDmode (VESSEL *vessel, int mode)
{
	oapiTriggerPanelRedrawArea (0, AID_HUDMODE);
}

DLLCLBK void ovcMFDmode (VESSEL *vessel, int mfd, int mode)
{
	switch (mfd) {
	case MFD_LEFT:
		oapiTriggerPanelRedrawArea (0, AID_MFD1_LBUTTONS);
		oapiTriggerPanelRedrawArea (0, AID_MFD1_RBUTTONS);
		break;
	case MFD_RIGHT:
		oapiTriggerPanelRedrawArea (0, AID_MFD2_LBUTTONS);
		oapiTriggerPanelRedrawArea (0, AID_MFD2_RBUTTONS);
		break;
	}
}

DLLCLBK bool ovcLoadPanel (VESSEL *vessel, int id)
{
	int xofs;
	DeltaGlider *dg = (DeltaGlider*)vessel;
	dg->ReleaseSurfaces();

	HBITMAP hBmp = LoadBitmap (g_Param.hDLL, MAKEINTRESOURCE (IDB_PANEL1+id));
	MFDSPEC mfds_left  = {{227, 61, 517, 351}, 6, 6, 47, 41};
	MFDSPEC mfds_right = {{636, 61, 926, 351}, 6, 6, 47, 41};

	switch (id) {
	case 0: // main panel
		oapiRegisterPanelBackground (hBmp, PANEL_ATTACH_BOTTOM|PANEL_MOVEOUT_BOTTOM, 0xffffff);
		oapiSetPanelNeighbours (-1,-1,1,-1);
		oapiRegisterMFD (MFD_LEFT,  mfds_left);
		oapiRegisterMFD (MFD_RIGHT, mfds_right);
		oapiRegisterPanelArea (AID_PROPELLANTSTATUS,  _R(  46,  84,  147, 148), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_ENGINESTATUS,      _R(1000,  90, 1112, 256), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_MAINENGINECONTROL, _R( 994, 296, 1048, 364), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBPRESSED);
		oapiRegisterPanelArea (AID_HOVRENGINECONTROL, _R(1062, 296, 1115, 364), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBPRESSED);
		oapiRegisterPanelArea (AID_ATTITUDEMODE,      _R(  76, 195,  116, 239), PANEL_REDRAW_MOUSE, PANEL_MOUSE_DOWN);
		oapiRegisterPanelArea (AID_NAVMODE,           _R(  38, 247,  153, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_HUDMODE,           _R(  15,  18,  122,  33), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_AOAINSTR,          _R( 206, 413,  262, 469), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_SLIPINSTR,         _R( 305, 413,  361, 469), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_SLOPEINSTR,        _R( 414, 405,  455, 476), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_RANGEINSTR,        _R( 476, 422,  510, 467), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_HORIZON,           _R( 569, 394,  665, 490), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_IGNORE);

		xofs = 182; // left MFD
		oapiRegisterPanelArea (AID_MFD1_BBUTTONS,     _R( 55+xofs,359,325+xofs, 377), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN); // bottom button row
		oapiRegisterPanelArea (AID_MFD1_LBUTTONS,     _R(    xofs,100, 25+xofs, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND); // left button column
		oapiRegisterPanelArea (AID_MFD1_RBUTTONS,     _R(353+xofs,100,378+xofs, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND); // right button column

		xofs = 591; // right MFD
		oapiRegisterPanelArea (AID_MFD2_BBUTTONS,     _R( 55+xofs,359,325+xofs, 377), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN); // bottom button row
		oapiRegisterPanelArea (AID_MFD2_LBUTTONS,     _R(    xofs,100, 25+xofs, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND); // left button column
		oapiRegisterPanelArea (AID_MFD2_RBUTTONS,     _R(353+xofs,100,378+xofs, 323), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBPRESSED, PANEL_MAP_BACKGROUND); // right button column

		// load bitmaps
		dg->srf[0] = oapiCreateSurface (LOADBMP (IDB_DIAL1));
		dg->srf[1] = oapiCreateSurface (LOADBMP (IDB_NAVBUTTON));
		dg->srf[2] = oapiCreateSurface (LOADBMP (IDB_LIGHT1));
		dg->srf[3] = oapiCreateSurface (LOADBMP (IDB_HORIZON));   oapiSetSurfaceColourKey (dg->srf[3], 0);

		break;
	case 1: // overhead panel
		oapiRegisterPanelBackground (hBmp, PANEL_ATTACH_TOP|PANEL_MOVEOUT_TOP, 0xffffff);
		oapiSetPanelNeighbours (-1,-1,-1,0);
		oapiRegisterPanelArea (AID_GEARBUTTON,        _R( 530, 183, 555, 221), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_GEARINDICATOR,     _R( 489, 186, 518, 217), PANEL_REDRAW_USER, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_DOCKBUTTON,        _R( 595, 183, 620, 221), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN);
		oapiRegisterPanelArea (AID_DOCKINDICATOR,     _R( 632, 186, 661, 217), PANEL_REDRAW_USER, PANEL_MOUSE_IGNORE, PANEL_MAP_BACKGROUND);
		oapiRegisterPanelArea (AID_OUTERAIRLOCKBUTTON,_R( 481,  31, 506,  69), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN);

		// load bitmaps
		dg->srf[0] = oapiCreateSurface (LOADBMP (IDB_SWITCH1));
		dg->srf[1] = oapiCreateSurface (LOADBMP (IDB_INDICATOR));

		break;
	}
	return hBmp != NULL;
}

DLLCLBK bool ovcPanelRedrawEvent (VESSEL *vessel, int id, int event, SURFHANDLE surf)
{
	DeltaGlider *dg = (DeltaGlider*)vessel;
	int i;

	switch (id) {
	// panel 0 events:
	case AID_PROPELLANTSTATUS:
		dg->RedrawPanel_PropellantStatus (surf);
		return true;
	case AID_ENGINESTATUS:
		dg->RedrawPanel_EngineStatus (surf);
		return true;
	case AID_ATTITUDEMODE:
		oapiBlt (surf, dg->srf[0], 0, 0, dg->GetAttitudeMode()*40, 0, 40, 44);
		return true;
	case AID_NAVMODE:
		dg->RedrawPanel_Navmode (surf);
		return true;
	case AID_HUDMODE:
		if ((i = oapiGetHUDMode()) > HUD_NONE)
			oapiBlt (surf, dg->srf[2], i*29+6, 0, 7, 0, 7, 7);
		return true;
	case AID_AOAINSTR:
		dg->RedrawPanel_AOA (surf);
		return true;
	case AID_SLIPINSTR:
		dg->RedrawPanel_Slip (surf);
		return true;
	case AID_SLOPEINSTR:
		dg->RedrawPanel_Slope (surf);
		return true;
	case AID_RANGEINSTR:
		dg->RedrawPanel_Range (surf);
		return true;
	case AID_HORIZON:
		dg->RedrawPanel_Horizon (surf);
		return true;
	case AID_MFD1_LBUTTONS:
		dg->RedrawPanel_MFDButton (surf, MFD_LEFT, 0);
		return true;
	case AID_MFD1_RBUTTONS:
		dg->RedrawPanel_MFDButton (surf, MFD_LEFT, 1);
		return true;
	case AID_MFD2_LBUTTONS:
		dg->RedrawPanel_MFDButton (surf, MFD_RIGHT, 0);
		return true;
	case AID_MFD2_RBUTTONS:
		dg->RedrawPanel_MFDButton (surf, MFD_RIGHT, 1);
		return true;

	// panel 1 events:
	case AID_GEARBUTTON:
		oapiBlt (surf, dg->srf[0], 0, 0, dg->gear_status == DeltaGlider::DOOR_OPEN ||
			dg->gear_status == DeltaGlider::DOOR_OPENING ? 0 : 25, 0, 25, 38);
		return true;
	case AID_DOCKBUTTON:
		oapiBlt (surf, dg->srf[0], 0, 0, dg->nose_status == DeltaGlider::DOOR_OPEN ||
			dg->nose_status == DeltaGlider::DOOR_OPENING ? 50 : 75, 0, 25, 38);
		return true;
	case AID_GEARINDICATOR:
		switch (dg->gear_status) {
		case DeltaGlider::DOOR_CLOSED: oapiBlt (surf, dg->srf[1], 0,  0, 0,  0, 29, 5); break;
		case DeltaGlider::DOOR_OPEN:   oapiBlt (surf, dg->srf[1], 0, 26, 0,  5, 29, 5); break;
		default:                       oapiBlt (surf, dg->srf[1], 0, 13, 0, 20, 29, 5); break;
		}
		return true;
	case AID_DOCKINDICATOR:
		switch (dg->nose_status) {
		case DeltaGlider::DOOR_CLOSED: oapiBlt (surf, dg->srf[1], 0,  0, 0, 10, 29, 5); break;
		case DeltaGlider::DOOR_OPEN:   oapiBlt (surf, dg->srf[1], 0, 26, 0, 15, 29, 5); break;
		default:                       oapiBlt (surf, dg->srf[1], 0, 13, 0, 20, 29, 5); break;
		}
		return true;
	case AID_OUTERAIRLOCKBUTTON:
		oapiBlt (surf, dg->srf[0], 0, 0, dg->lock_status == DeltaGlider::DOOR_OPEN ||
			dg->lock_status == DeltaGlider::DOOR_OPENING ? 0 : 25, 0, 25, 38);
		return true;
	}
	return false;
}

DLLCLBK bool ovcPanelMouseEvent (VESSEL *vessel, int id, int event, int mx, int my)
{
	DeltaGlider *dg = (DeltaGlider*)vessel;
	int mfd;

	switch (id) {
	// panel 0 events:
	case AID_MAINENGINECONTROL:
		if (mx < 25) {
			if (my < 33)
				dg->IncEngineLevel (ENGINE_MAIN, oapiGetSimStep() *  0.5);
			else if (my >= 36)
				dg->IncEngineLevel (ENGINE_MAIN, oapiGetSimStep() * -0.5);
		} else if (mx >= 29) {
			if (my < 20)
				dg->SetEngineLevel (ENGINE_MAIN, 1.0);
			else if (my >= 24 && my < 44)
				dg->SetEngineLevel (ENGINE_MAIN, 0.0);
			else if (my >= 48)
				dg->SetEngineLevel (ENGINE_MAIN, -1.0);
		}
		return true;
	case AID_HOVRENGINECONTROL:
		if (mx < 25) {
			if (my < 33)
				dg->IncEngineLevel (ENGINE_HOVER, oapiGetSimStep() * 0.5);
			else if (my >= 36)
				dg->IncEngineLevel (ENGINE_HOVER, oapiGetSimStep() * -0.5);
		} else if (mx >= 29) {
			if (my < 33)
				dg->SetEngineLevel (ENGINE_HOVER, 1.0);
			else if (my >= 36)
				dg->SetEngineLevel (ENGINE_HOVER, 0.0);
		}
		return true;
	case AID_ATTITUDEMODE:
		if (event & PANEL_MOUSE_LBDOWN) return dg->DecAttMode();
		if (event & PANEL_MOUSE_RBDOWN) return dg->IncAttMode();
		return false;
	case AID_NAVMODE:
		dg->ToggleNavmode ((mx/39)*2 + my/39 + 1);
		return true;
	case AID_HUDMODE:
		if (mx%29 < 20) oapiSetHUDMode (HUD_NONE+(mx/29));
		return true;
	case AID_MFD1_BBUTTONS:
	case AID_MFD2_BBUTTONS:
		mfd = (id == AID_MFD1_BBUTTONS ? MFD_LEFT : MFD_RIGHT);
		if (mx < 26)                    oapiToggleMFD_on (mfd);
		else if (mx >= 214 && mx < 240) oapiSendMFDKey (mfd, OAPI_KEY_F1);
		else if (mx > 244)              oapiSendMFDKey (mfd, OAPI_KEY_GRAVE);
		return true;
	case AID_MFD1_LBUTTONS:
	case AID_MFD1_RBUTTONS:
	case AID_MFD2_LBUTTONS:
	case AID_MFD2_RBUTTONS:
		if (my%41 < 18) {
			mfd = (id <= AID_MFD1_RBUTTONS ? MFD_LEFT : MFD_RIGHT);
			int bt = my/41 + (id == AID_MFD1_LBUTTONS || id == AID_MFD2_LBUTTONS ? 0 : 6);
			oapiProcessMFDButton (mfd, bt, event);
			return true;
		}
		break;
	// panel 1 events:
	case AID_GEARBUTTON:
		dg->ActivateLandingGear (my < 19 ? DeltaGlider::DOOR_CLOSING : DeltaGlider::DOOR_OPENING);
		break;
	case AID_DOCKBUTTON:
		dg->ActivateDockingPort (my < 19 ? DeltaGlider::DOOR_CLOSING : DeltaGlider::DOOR_OPENING);
		break;
	case AID_OUTERAIRLOCKBUTTON:
		dg->ActivateAirlock (my < 19 ? DeltaGlider::DOOR_CLOSING : DeltaGlider::DOOR_OPENING);
		break;
	}
	return false;
}

DLLCLBK int ovcConsumeBufferedKey (VESSEL *vessel, DWORD key, bool down, char *kstate)
{
	if (!down) return 0; // only process keydown events

	DeltaGlider *dg = (DeltaGlider*)vessel;

	if (KEYMOD_SHIFT (kstate)) {
	} else if (KEYMOD_CONTROL (kstate)) {
		switch (key) {
		case OAPI_KEY_DIVIDE:
			if (vessel->SetAttitudeMode (vessel->GetAttitudeMode() >= 1 ? 0 : 1))
				oapiTriggerPanelRedrawArea (0, AID_ATTITUDEMODE);
			return 1;
		}
	} else {
		switch (key) {
		case OAPI_KEY_G:  // "operate landing gear"
			dg->RevertLandingGear ();
			return 1;
		case OAPI_KEY_K:  // "operate docking port"
			dg->RevertDockingPort ();
			return 1;
		case OAPI_KEY_O:  // "operate outer airlock"
			dg->RevertAirlock ();
			return 1;
		case OAPI_KEY_DIVIDE:
			if (vessel->ToggleAttitudeMode()) oapiTriggerPanelRedrawArea (0, AID_ATTITUDEMODE);
			return 1;
		}
	}
	return 0;
}
