// ==============================================================
//                ORBITER MODULE: DeltaGlider
//                  Part of the ORBITER SDK
//            Copyright (C) 2001 Martin Schweiger
//                   All rights reserved
//
// DeltaGlider.h
// Class interface for delta glider vessel class module
// ==============================================================

#ifndef __DELTAGLIDER_H
#define __DELTAGLIDER_H

#include "orbitersdk.h"
#include "resource.h"

// ==========================================================
// Some vessel class caps
// ==========================================================

const double MAX_MAIN_THRUST = 2.16e5;
// Main engine max thust [N]

const double MAX_RETRO_THRUST = 5.4e4;
// Retro engine max thrust [N]

const double MAX_HOVER_THRUST = 2.16e5;
// Hover engine max thrust [N]

const double MAX_ATTITUDE_THRUST = 5e3;
// Attitude control system max thrust [N]

const double GEAR_OPERATING_SPEED = 0.15;
// Opening/closing speed of landing gear (1/sec)
// => gear cycle ~ 6.7 sec

const double NOSE_OPERATING_SPEED = 0.1;
// Opening/closing speed of nose cone docking mechanism (1/sec)
// => cycle = 10 sec

const double AIRLOCK_OPERATING_SPEED = 0.1;
// Opening/closing speed of outer airlock (1/sec)
// => cycle = 10 sec

const int nsurf = 4; // number of bitmap handles

// ==========================================================
// Interface for derived vessel class: DeltaGlider
// ==========================================================

class DeltaGlider: public VESSEL {
public:
	DeltaGlider (OBJHANDLE hObj, int fmodel);
	~DeltaGlider ();
	void LoadState (FILEHANDLE scn, VESSELSTATUS *vs);
	void SaveState (FILEHANDLE scn);
	void SetClassCaps (FILEHANDLE cfg);
	void DefineAnimations ();
	void ReleaseSurfaces();
	void Timestep (double simt);
	void DrawNeedle (HDC hDC, int x, int w, double rad, double angle, double *pangle = 0, double speed = PI);
	void RedrawPanel_PropellantStatus (SURFHANDLE surf);
	void RedrawPanel_EngineStatus (SURFHANDLE surf);
	void RedrawPanel_AOA (SURFHANDLE surf);
	void RedrawPanel_Slip (SURFHANDLE surf);
	void RedrawPanel_Slope (SURFHANDLE surf);
	void RedrawPanel_Range (SURFHANDLE surf);
	void RedrawPanel_Horizon (SURFHANDLE surf);
	void RedrawPanel_Navmode (SURFHANDLE surf);
	void RedrawPanel_MFDButton (SURFHANDLE surf, int mfd, int side);
	bool DecAttMode ();
	bool IncAttMode ();

	double alt;       // altitude
	double range;     // glide range
	double aoa_ind;   // angle of AOA needle (NOT AOA!)
	double slip_ind;  // angle of slip indicator needle
	double slope, slope_ind; // glide slope and angle of slope indicator needle
	DWORD navmode_bt_status; // navmode button status

	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } nose_status, gear_status, lock_status;
	void ActivateLandingGear (DoorStatus action);
	void ActivateDockingPort (DoorStatus action);
	void ActivateAirlock (DoorStatus action);
	void RevertLandingGear ();
	void RevertDockingPort ();
	void RevertAirlock ();
	double nose_proc, gear_proc, lock_proc;     // logical status
	UINT anim_gear;     // handle for landing gear animation
	UINT anim_nose;     // handle for nose cone animation
	UINT anim_lock;     // handle for airlock animation

	SURFHANDLE srf[nsurf];  // handles for panel bitmaps
};

typedef struct {
	HINSTANCE hDLL;
	HFONT font[2];
	DWORD col[4];
	HBRUSH brush[4];
	HPEN pen[2];
} GDIParams;

// Panel area identifiers:
// Panel 0
#define AID_PROPELLANTSTATUS  0
#define AID_ENGINESTATUS      1
#define AID_MAINENGINECONTROL 2
#define AID_HOVRENGINECONTROL 3
#define AID_ATTITUDEMODE      4
#define AID_NAVMODE           5
#define AID_HUDMODE           6
#define AID_AOAINSTR          7
#define AID_SLIPINSTR         8
#define AID_SLOPEINSTR        9
#define AID_RANGEINSTR       10
#define AID_HORIZON          11
#define AID_MFD1_BBUTTONS    12
#define AID_MFD1_LBUTTONS    13
#define AID_MFD1_RBUTTONS    14
#define AID_MFD2_BBUTTONS    15
#define AID_MFD2_LBUTTONS    16
#define AID_MFD2_RBUTTONS    17

// Panel 1
#define AID_GEARBUTTON         100
#define AID_GEARINDICATOR      101
#define AID_DOCKBUTTON         102
#define AID_DOCKINDICATOR      103
#define AID_OUTERAIRLOCKBUTTON 104

#endif // !__DELTAGLIDER_H